#include <xc.h>

// #pragma config statements should precede project file includes

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config PWRTE = ON       // Power-up Timer Enable (PWRT enabled)
#pragma config MCLRE = OFF      // MCLR Pin Function Select (MCLR/VPP pin function is digital input)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled. SBOREN bit is ignored.)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config STVREN = OFF     // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will not cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), 1.9V trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)

// project includes
#include "global.h"
#include "nrf24l01.h"

// global variables
SensorData g_pData;                     // tx/rx sensor and config data

uint8_t SendSensor()
{
    // send sensor readings via nrf24l01+
    SensorData pEcho;
    uint8_t bConfigChanged = 0;

    // when powered down, 1.5ms settling is
    // required before it can enter RX or TX
    NRF_TXMode();
    __delay_ms(2);

    // transmit first
    NRF_TXData((uint8_t*)&g_pData, sizeof(g_pData));

    // check for echo
    NRF_RXMode();
    __delay_ms(50);
    if (NRF_DataAvailable())
    {
        // read echo
        NRF_RXData((uint8_t*)&pEcho, sizeof(pEcho));

        // check for changed config
        if (pEcho.uId == g_pData.uId && pEcho.uConfig != g_pData.uConfig)
        {
            // config has changed - save it
            g_pData.uConfig = pEcho.uConfig;
            bConfigChanged = 1;
        }
    }

    // power down the transceiver
    NRF_PowerDown();

    // return true if config changed
    return bConfigChanged;
}

void ReadSensor()
{
    // read sensor and return results in global variables
    t_short pResult;

    // switch sensor on in one direction and let it settle
    LATA2 = 0;
    LATA5 = 1;
    __delay_ms(10);

    // sensor is on RA4, which is channel 10
    ADCON0bits.CHS = 0b01010;

    // enable ADC
    ADCON0bits.AD1ON = 1;

    // acquisition time wait
    __delay_us(5);

    // start conversion and
    // wait until complete
    ADCON0bits.ADGO = 1;
    while(ADCON0bits.ADGO);

    // disable ADC
    ADCON0bits.AD1ON = 0;

    // capture first result
    pResult.uMSB = AD1RES0H & 0b11;
    pResult.uLSB = AD1RES0L;
    g_pData.nReading1 = pResult.nShort;

    // switch sensor on in reverse direction and let it settle
    LATA5 = 0;
    LATA2 = 1;
    __delay_ms(10);

    // enable ADC
    ADCON0bits.AD1ON = 1;

    // acquisition time wait
    __delay_us(5);

    // start conversion and
    // wait until complete
    ADCON0bits.ADGO = 1;
    while(ADCON0bits.ADGO);

    // disable ADC
    ADCON0bits.AD1ON = 0;

    // capture second result
    pResult.uMSB = AD1RES0H & 0b11;
    pResult.uLSB = AD1RES0L;
    g_pData.nReading2 = pResult.nShort;

    // switch sensor off
    LATA2 = 0; 
    LATA5 = 0;
}

void main()
{
    // 4x PLL, 8MHz internal oscillator = 32MHz clock
    OSCCON = 0xF0;

    // wait for HIFINTOSC to stabilise
    while (OSCSTATbits.HFIOFR == 0);
    while (OSCSTATbits.HFIOFS == 0);

    // configure alternate pin functions
    RXDTSEL = 0;                // RX on RC5 not RA4 (RC5 unused)
    SDOSEL = 0;                 // SDO on RC2 not RA4
    SSSEL = 1;                  // SS on RC3 not RA3 (RC3 unused for SPI master)
    SDSEL = 0;                  // SDI on RC1 not RA3
    TXCKSEL = 0;                // TX on RC4 not RC3 (RC4 unused)

    // configure PORTA inputs and outputs
    TRISA2 = 0;                 // RA2 is sensor power 1
    TRISA3 = 1;                 // RA3 unused (VPP) but is input-only
    TRISA4 = 1;                 // RA4 is analogue input
    TRISA5 = 0;                 // RA5 is sensor power 2

    // unused PORTA pins are set to outputs and are
    // held LOW to minimise power loss when sleeping
    TRISA0 = 0;                 // RA0 unused (ICSPDAT)
    LATA0 = 0;
    TRISA1 = 0;                 // RA1 unused (ICSPCLK)
    LATA1 = 0;

    // RA2/RA5 provide power to the sensor
    // so initially turn the power off
    LATA2 = 0;
    LATA5 = 0;

    // mostly digital, disable analogue
    // except on RA4 for ADC reading
    ANSELA = 0x00;
    ANSELC = 0x00;
    ANSA4 = 1;

    // configure ADC
    ADCON1bits.ADCS = 0b011;    // FRC = 1 to 6uS @ 32MHz
    ADCON1bits.ADFM = 1;        // right justified result
    ADCON1bits.ADPREF = 0b00;   // use Vdd as Vref

    // configure PORTC inputs and outputs
    TRISC0 = 0;                 // SCK is output
    TRISC1 = 1;                 // SDI is input
    TRISC2 = 0;                 // SDO is output

    // unused PORTC pins are set to outputs and are
    // held LOW to minimise power loss when sleeping
    TRISC3 = 1;                 // SS unused
    LATC3 = 0;
    TRISC4 = 0;                 // TX unused
    LATC4 = 0;
    TRISC5 = 0;                 // RX unused
    LATC5 = 0;

    // initialise sensor data
    g_pData.uId = SENSOR_ID;            // sensor id
    g_pData.uConfig = DFT_CONFIG;       // initial reporting interval
    g_pData.uSequence = 0;              // sequence number
    
    // initialise RF
    NRF_Setup();

    // initialise additional watchdog scaler
    uint16_t uWatchdogCounter = 0;

    // configure watchdog interval
    WDTCONbits.WDTPS = (g_pData.uConfig & 0b11111);

    // main loop
    while (1)
    {
        // disable watchdog while we do this bit
        WDTCONbits.SWDTEN = 0;

        // if it's time for a reading ...
        if (uWatchdogCounter == 0)
        {
            // ... if we didn't find the RF module ...
            if (!NRF_IsInitialised())
            {
                // ... attempt to reinitialise
                NRF_Setup();
            }

            // take a reading and send via nrf24l01+
            ReadSensor();
            if (SendSensor())
            {
                // config has changed so set watchdog interval
                WDTCONbits.WDTPS = (g_pData.uConfig & 0b11111);
            }

            // increment rolling sequence
            ++g_pData.uSequence;

            // reset watchdog scaler
            uWatchdogCounter = WATCHDOG_SCALE;
        }
        else
        {
            // decrement scaler until next wake-up
            --uWatchdogCounter;
        }

        // put watchdog back on guard
        WDTCONbits.SWDTEN = 1;

        // go to sleep until watchdog wakes
        SLEEP();
    }
}
