///////////////////////////////////////////////////////////////////////////////
//
// File:    nrf24l01.c
// Author:  Roger Rowland
//
// Created: December 2014
//
///////////////////////////////////////////////////////////////////////////////

#include <xc.h>
#include "global.h"
#include "nRF24L01.h"

///////////////////////////////////////////////////////////////////////////////
// SPI(nRF24L01) commands
///////////////////////////////////////////////////////////////////////////////

#define READ_REG        0x00  // Define read command to register
#define WRITE_REG       0x20  // Define write command to register
#define RD_RX_PLOAD     0x61  // Define RX payload register address
#define WR_TX_PLOAD     0xA0  // Define TX payload register address
#define FLUSH_TX        0xE1  // Define flush TX register command
#define FLUSH_RX        0xE2  // Define flush RX register command
#define REUSE_TX_PL     0xE3  // Define reuse TX payload register command
#define NRF_NOP         0xFF  // Define No Operation, might be used to read status register

///////////////////////////////////////////////////////////////////////////////
// SPI(nRF24L01) registers(addresses)
///////////////////////////////////////////////////////////////////////////////

#define NRF_CONFIG      0x00  // 'Config' register address
#define EN_AA           0x01  // 'Enable Auto Acknowledgment' register address
#define EN_RXADDR       0x02  // 'Enabled RX addresses' register address
#define SETUP_AW        0x03  // 'Setup address width' register address
#define SETUP_RETR      0x04  // 'Setup Auto. Retrans' register address
#define RF_CH           0x05  // 'RF channel' register address
#define RF_SETUP        0x06  // 'RF setup' register address
#define NRF_STATUS      0x07  // 'Status' register address
#define OBSERVE_TX      0x08  // 'Observe TX' register address
#define CD              0x09  // 'Carrier Detect' register address
#define RX_ADDR_P0      0x0A  // 'RX address pipe0' register address
#define RX_ADDR_P1      0x0B  // 'RX address pipe1' register address
#define RX_ADDR_P2      0x0C  // 'RX address pipe2' register address
#define RX_ADDR_P3      0x0D  // 'RX address pipe3' register address
#define RX_ADDR_P4      0x0E  // 'RX address pipe4' register address
#define RX_ADDR_P5      0x0F  // 'RX address pipe5' register address
#define TX_ADDR         0x10  // 'TX address' register address
#define RX_PW_P0        0x11  // 'RX payload width, pipe0' register address
#define RX_PW_P1        0x12  // 'RX payload width, pipe1' register address
#define RX_PW_P2        0x13  // 'RX payload width, pipe2' register address
#define RX_PW_P3        0x14  // 'RX payload width, pipe3' register address
#define RX_PW_P4        0x15  // 'RX payload width, pipe4' register address
#define RX_PW_P5        0x16  // 'RX payload width, pipe5' register address
#define FIFO_STATUS     0x17  // 'FIFO Status Register' register address

///////////////////////////////////////////////////////////////////////////////
// local data
///////////////////////////////////////////////////////////////////////////////

#define ADDR_SIZE       5                       // number of address bytes
static const uint8_t RXTX_ADDR[6][ADDR_SIZE] =  // randomly chosen addresses
{
    {0xE1, 0xA3, 0xB0, 0xD5, 0xF1},
    {0xD1, 0xB7, 0xA4, 0x09, 0xAB},
    {0xD2, 0xB7, 0xA4, 0x09, 0xAB},
    {0xD3, 0xB7, 0xA4, 0x09, 0xAB},
    {0xD4, 0xB7, 0xA4, 0x09, 0xAB},
    {0xD5, 0xB7, 0xA4, 0x09, 0xAB}
};

uint8_t g_bModulePresent = 0;                   // non-zero if rf module present

///////////////////////////////////////////////////////////////////////////////
// internally used function prototypes
///////////////////////////////////////////////////////////////////////////////

void SPIInit();                                 // initialise SPI
uint8_t SPITransfer(uint8_t uData);             // transfer one byte
uint8_t ReadRegister(uint8_t reg);              // read a register
uint8_t ReadStatus();                           // read and return status
void WriteRegister(uint8_t reg, uint8_t val);   // write a register
void WriteAddress(uint8_t reg, uint8_t num,     // write an address
        const uint8_t* addr);
void WriteCommand(uint8_t command);             // write a command
void FlushTXRX();                               // flush tx and rx buffers

///////////////////////////////////////////////////////////////////////////////
// public function implementations
///////////////////////////////////////////////////////////////////////////////

void NRF_Setup()
{
    // initialise nRF24l01+
    g_bModulePresent = 0;
    
    // start SPI
    SPIInit();

    // nRF24l01+ needs 100ms max for power up
    __delay_ms(100);

    // initalise the CSN and CE pins
    NRF_CSN_TRIS = 0;
    NRF_CE_TRIS = 0;

    // keep CSN high by default because all SPI commands must
    // be started by a high to low transition on CSN. Leave
    // the chip disabled initially by setting CE low
    NRF_CSN = 1;
    NRF_CE = 0;

    WriteRegister(NRF_CONFIG, 0x0B);                // 8-bit CRC, POWER UP, PRX
    WriteRegister(EN_AA, 0x00);                     // disable auto ack
    WriteRegister(EN_RXADDR, 0x01);                 // enable data pipe 0
    WriteRegister(SETUP_AW, 0x03);                  // 5-byte addresses
    WriteRegister(SETUP_RETR, 0x00);                // retransmit disabled
    WriteRegister(RF_CH, 0x4C);                     // randomly chosen RF channel (76)
    WriteRegister(RF_SETUP, 0x27);                  // 250kbps, 0dBm
    WriteRegister(RX_PW_P0, sizeof(SensorData));    // RX payload

    // set rx/tx addresses
    WriteAddress(RX_ADDR_P0, ADDR_SIZE, RXTX_ADDR[SENSOR_ID - 1]);
    WriteAddress(TX_ADDR, ADDR_SIZE, RXTX_ADDR[SENSOR_ID - 1]);

    // flush fifos before we start
    FlushTXRX();

    // make sure we have a real module
    if ((ReadRegister(NRF_CONFIG) & 0x08) != 0)
        g_bModulePresent = 1;
}

uint8_t NRF_IsInitialised()
{
    // return non-zero if initialised
    return g_bModulePresent;
}

void NRF_RXMode()
{
    // go to receive mode
    WriteRegister(NRF_CONFIG, 0x0B); // 8-bit CRC, POWER UP, PRX
    NRF_CE = 1;
}

void NRF_TXMode()
{
    // go to transmit mode
    NRF_CE = 0;
    WriteRegister(NRF_CONFIG, 0x0A); // 8-bit CRC, POWER UP, PTX
}

void NRF_PowerDown()
{
    // low-power mode
    NRF_CE = 0;
    WriteRegister(NRF_CONFIG, 0);
}

void NRF_RXData(uint8_t *pData, uint8_t uBytes)
{
    // read received data via SPI, noting that each
    // SPI command must start with high-low on CSN
    NRF_CSN = 0;

    // read payload from fifo
    SPITransfer(RD_RX_PLOAD);
    for (uint8_t i = 0; i < uBytes; ++i)
    {
        // dummy transfer to read a byte
        pData[i] = SPITransfer(0);
    }

    // leave CSN high in preparation for next command
    NRF_CSN = 1;

    // clear status bit
    WriteRegister(NRF_STATUS, 0x40);
}

void NRF_TXData(uint8_t *pData, uint8_t uBytes)
{
    // transmit data via SPI, noting that each
    // SPI command must start with high-low on CSN
    NRF_CSN = 0;

    // write payload to fifo
    SPITransfer(WR_TX_PLOAD);
    for (uint8_t i = 0; i < uBytes; ++i)
    {
        // write a data byte
        SPITransfer(pData[i]);
    }

    // leave CSN high in preparation for next command
    NRF_CSN = 1;

    // pulse for Thce (10us) + a bit (2us)
    NRF_CE = 1;
    __delay_us(12);
    NRF_CE = 0;

    // if we have a real rf module ...
    if (g_bModulePresent)
    {
        // ... wait for data to be sent
        uint8_t uStat;
        do
        {
            // poll status
            uStat = ReadStatus();
        } while ((uStat & 0x20) == 0);
    }

    // clear status bit
    WriteRegister(NRF_STATUS, 0x20);
}

uint8_t NRF_DataAvailable()
{
    // return non-zero if data is waiting
    uint8_t stat = ReadStatus();
    return (stat & 0x40) != 0;
}

///////////////////////////////////////////////////////////////////////////////
// internal function implementations
///////////////////////////////////////////////////////////////////////////////

void SPIInit()
{
    // initialise SPI
    SSPEN = 0;              // disable spi
    CKE = 1;                // transmit on active-idle clock transition
    SSPCON1 = 0x01;         // CKP = 0, SCK = Fosc/16 (1MHz)
    SMP = 1;                // input sampled at end of output
    PIR1bits.SSP1IF = 0;    // clear interrupt flag
    SSPEN = 1;              // enable spi
}

uint8_t SPITransfer(uint8_t uData)
{
    // transfer one byte
    SSPBUF = uData;
    while (!PIR1bits.SSP1IF);
    PIR1bits.SSP1IF = 0;
    return SSPBUF;
}

void FlushTXRX()
{
    // flush rx and tx buffers
    WriteRegister(NRF_STATUS, 0x70);
    WriteCommand(FLUSH_RX);
    WriteCommand(FLUSH_TX);
}

void WriteRegister(uint8_t reg, uint8_t val)
{
    // write to a register
    NRF_CSN = 0;
    SPITransfer(WRITE_REG | reg);
    SPITransfer(val);
    NRF_CSN = 1;
}

void WriteAddress(uint8_t reg, uint8_t num, const uint8_t* addr)
{
    // write an address (normally 3 to 5 bytes)
    NRF_CSN = 0;
    SPITransfer(WRITE_REG | reg);
    for (uint8_t i = 0; i < num; i++)
        SPITransfer(addr[i]);
    NRF_CSN = 1;
}

uint8_t ReadRegister(uint8_t reg)
{
    // read a register
    NRF_CSN = 0;
    SPITransfer(READ_REG | reg);
    uint8_t val = SPITransfer(0x00);
    NRF_CSN = 1;
    return val;
}

uint8_t ReadStatus()
{
    // write nop to read status
    NRF_CSN = 0;
    uint8_t val = SPITransfer(NRF_NOP);
    NRF_CSN = 1;
    return val;
}

void WriteCommand(uint8_t command)
{
    // write a command
    NRF_CSN = 0;
    SPITransfer(command);
    NRF_CSN = 1;
}
